/*
  @file   CBetaDriver.cpp
  @brief  wK_CuBetaRNA ̃hCoNX
  @author AdIn Research,Inc
  @date   2007/02/22
*/

#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "rnaZ0usr.h"
#include "rnaB0usr.h"
#include "rnaS1def.h"
#include "rnaF0usr.h"
#include "CBetaDriver.h"


#define BD_PARAM_VALUE  "> 1000\nML WEIGHT NO_CLR_H EXTEND NO_ACT\nDG 0.5 0.5 0.5 0.5 0.0 3.0 0.5\nDH 10.0\nDA 1.0\nXB 50.0 0.0\nXV 40.0\nXW 50.0\nWB 80.0 10.0\nWG 10.0 10.0\nE\n"



/*
  @brief      AwKpobt@g
  @return     bool
  @author     AdIn Research,Inc
  @date       2007/02/22
*/
bool CBetaDriver::expandBuf( void )
{
  size_t     tSize;
  tOneDatas* pNewBuf;

  // AwKpobt@ɋ󂫂L̂ŁAgKv͂Ȃ
  if( m_iBufSize > m_iBufUse )
    return true;

  // gɕKvȃTCYvZ
  tSize = sizeof( tOneDatas ) * ( m_iBufSize + BD_EXPAND_SIZE );

  // gς݂̘AwKpobt@쐬
  pNewBuf = (tOneDatas*)malloc( tSize );
  if( ! pNewBuf )
    return false;

  // AwKpobt@
  memset( (void*)pNewBuf, 0x00, tSize );

  // gO̘AwKpobt@Ag̘AwKpobt@Ɉړ
  if( m_iBufSize ) {
    tSize = sizeof( tOneDatas ) * m_iBufSize;
    memcpy( (void*)pNewBuf, (void*)m_pBuf, tSize );
    free( (void*)m_pBuf );
  }

  // AwKpobt@XV
  m_pBuf = pNewBuf;
  m_iBufSize += BD_EXPAND_SIZE;

  return true;
}


/*
  @brief      AwKpobt@NA
  @author     AdIn Research,Inc
  @date       2007/02/22
*/
void CBetaDriver::clearBuf( void )
{
  int i;

  // AwKpobt@ɂ܂ʃf[^J
  for( i = 0; i < m_iBufUse; i++ )
    free( (void*)( m_pBuf[i].pdData ) );

  // AwKpobt@J
  if( m_iBufSize )
    free( (void*)m_pBuf );

  m_pBuf = NULL;
  m_iBufSize = m_iBufUse = 0;
}


/*
  @brief      ʃf[^AwKt@Cɕۑ
  @param[in]  wKɎgpʃf[^Zbg̐
  @param[in]  ʃf[^Zbg
  @return     bool
  @author     AdIn Research,Inc
  @date       2007/02/22
*/
bool CBetaDriver::outDataFile( int iInput, tOneDatas* pInput )
{
  FILE* fp;
  int   i, j;
  bool  bStudy;

  // ݂̃[hiwK[hA_[hHj擾
  if( pInput[0].iOutputID >= 0 )
    bStudy = true;
  else
    bStudy = false;

  // wKf[^t@CJ
  fp = fopen( m_sInputPath, "wt" );
  if( ! fp )
    return false;

  // ʍڐAʑΏۂ̍ő吔ݒ
  fprintf( fp, "H %d %d\n", m_iInputCh, BD_OUTPUT_MAX );

  // wK[ĥƂɕKvȋLq쐬
  if( bStudy ) {
    for( i = 0; i < m_iInputCh; i++ )
      fprintf( fp, "I%d In%d\t0.0\t1.0\n", i, i );

    for( i = 0; i < BD_OUTPUT_MAX; i++ )
      fprintf( fp, "O%d Out%d\n", i, i );

    fprintf( fp, "S" );

    for( i = 0; i < m_iInputCh; i++ )
      fprintf( fp, " 1" );

    fprintf( fp, "\n" );
  }

  // eʃf[^o
  fprintf( fp, "#\n" );
  for( i = 0; i < iInput; i++ ) {
    // ʂ̍ڒlo
    fprintf( fp, "D" );
    for( j = 0; j < m_iInputCh; j++ )
      fprintf( fp, "\t%lf", pInput[i].pdData[j] );

    // ʂ̍ڒl̋؂o
    if( bStudy )
      fprintf( fp, "\t%d\n", pInput[i].iOutputID );
    else
      fprintf( fp, "\n" );
  }
  fprintf( fp, "#\nE\n" );

  // ㏈
  fclose( fp );

  return true;
}


/*
  @brief      wKs
  @return     bool
  @author     AdIn Research,Inc
  @date       2007/02/22
*/
bool CBetaDriver::study( void )
{
    Ferr            err;
    STREAM          stream;
    RS1DATA        *prs1;
    unsigned long   ulStat, ulMode, ulCycle = 0L;
    static RLMFP    rlm;
    static RLEXT    rle;
    static RLWEIGHT rlw;
    RANA            raSysMax;

  /*---------------------------- Clear Data -------------------------------*/
  ulStat = ulMode = 0L;
  memset(&rlm, 0, sizeof(RLMFP));
  memset(&rle, 0, sizeof(RLEXT));
  memset(&rlw, 0, sizeof(RLWEIGHT));
  /*------------------------- Set Default Parameter -----------------------*/
  rsSetLParaDefaults(RST_RS1, &ulStat, &rlm, &rle, &rlw, &raSysMax);
  /*------------------------ Load Parameter File --------------------------*/
  if (FENOER != (err = rsLoadLParaFile( m_sParamPath,
                               (int *)NULL, (RNUM *)NULL,
                               &ulStat, &ulMode, &rlm, &rle, &rlw,
                               (unsigned long *)NULL, (unsigned long *)NULL,
                               (RLMFP *)NULL, (RLEXT *)NULL, (RLWEIGHT *)NULL,
                               &raSysMax, (RMD*) NULL)) )
    goto f_err;

  /*--------------------------- Initialize --------------------------------*/
  if ( m_bStudyOnly ) {
    err = RS1LibInit(m_sInputPath, ulMode, ulStat, &rlm, &rle, &rlw,
                     raSysMax, &prs1, &stream);
  }
  else {
    ulStat |= MRL_LEARN;
    err = RS1LibSet(m_sInputPath, UNDEFINED_ID,
                    ulMode, ulStat, &rlm, &rle, &rlw, &prs1, &stream);
  }
  if (FENOER != err)
    goto f_err;
  /*-----------------------------------------------------------------------*/
  for (;;) {
    RID riTeach;
    /*------------------------- Execute Learning ------------------------*/
    if (FENOER != (err = RS1LibExec(prs1, &riTeach, &stream))) {
      if (err == F0IOER)
        break;
      else if (err != F0RNGE)
        goto f_err;
    }
    if (FENOER != (err = RS1_PrintLearnResult(++ulCycle, prs1, riTeach,
                                    stdout, m_iPrintMode, err)) )
      goto f_err;
  }
  /*----------------------------- Ferr Sim. Data --------------------------*/
  if (FENOER != (err = RS1LibFree(prs1))) {
    fprintf(stderr, "Error : memory garbled (%d)\n", (int)err);
    goto f_err;
  }
  /*-----------------------------  Save study data ------------------------*/
  if (FENOER != (err = RCSaveData( m_sStudyPath ))) {
    fprintf(stderr, "Cannot save [%s](%d)\n", m_sStudyPath, (int)err);
    return false;
  }

  return true;

f_err:
  fprintf( stdout, "Error in RNA-study-process, err = %d\n", (int)err );
  RZPrintError(stderr, "RS1 error");

  return false;
}

/*
  @brief      _s
  @param[out] piRet _IDԍ
  @param[out] vx
  @return     bool
  @author     AdIn Research,Inc
  @date       2007/02/22
*/
bool CBetaDriver::detect( int* piRet, double* pdTrust )
{
  Ferr            err;
  STREAM          stream;
  RS1DATA        *prs1;
  unsigned long   ulStat, ulMode, ulCycle = 0L;
  static RLMFP    rlm;
  static RLEXT    rle;
  static RLWEIGHT rlw;
  RANA            raSysMax;

  /*---------------------------- Clear Data -------------------------------*/
  ulStat = ulMode = 0L;
  memset(&rlm, 0, sizeof(RLMFP));
  memset(&rlm, 0, sizeof(RLEXT));
  memset(&rlm, 0, sizeof(RLWEIGHT));
  /*------------------------- Set Default Parameter -----------------------*/
  rsSetLParaDefaults(RST_RS1, &ulStat, &rlm, &rle, &rlw, &raSysMax);
  /*------------------------ Load Parameter File --------------------------*/
  if (FENOER != (err = rsLoadLParaFile(m_sParamPath,
                               (int *)NULL, (RNUM *)NULL,
                               &ulStat, &ulMode, &rlm, &rle, &rlw,
                               (unsigned long *)NULL, (unsigned long *)NULL,
                               (RLMFP *)NULL, (RLEXT *)NULL, (RLWEIGHT *)NULL,
                               &raSysMax, (RMD*) NULL)) )
    goto f_err;
  /*--------------------------- Initialize --------------------------------*/
  if (FENOER != (err = RS1LibSet(m_sInputPath, UNDEFINED_ID,
                                   ulMode, ulStat,
                                   &rlm, &rle, &rlw, &prs1, &stream)) ) {
    if (F0RNGE == err)
      fputs("\n Warning: Find inconsitency in Normalize Table !!\n\n", stderr);
    else
      goto f_err;
  }
  /*-----------------------------------------------------------------------*/
  *piRet = -1;
  for (;;) {
    RMD rgMax;
    int iMax;
     /*------------------------- Execute Learning ------------------------*/
    if (FENOER != (err = RS1LibProd(prs1, &iMax, &rgMax, &stream))) {
      if (err == F0IOER)
        break;
      else if (err != F0RNGE)
        goto f_err;
    }

    if( *piRet < 0 ) {
      *piRet = iMax;
      *pdTrust = (double)rgMax;
    }

    if (FENOER != (err = RS1_PrintInferResult(++ulCycle, prs1, iMax, rgMax,
                                                stdout, m_iPrintMode, err)) )
      goto f_err;
  }

  /*----------------------------- Ferr Sim. Data --------------------------*/
  if (FENOER != (err = RS1LibFree(prs1))) {
    fprintf(stderr, "Error : memory garbled (%d)\n", (int)err);
    goto f_err;
  }

  return true;

f_err:
  fprintf( stdout, "Error in RNA-detect-process, err = %d\n", (int)err );
  RZPrintError(stderr, "RS1 error");
  return false;
}


/*
  @brief      RXgN^
  @author     AdIn Research,Inc
  @date       2007/02/22
*/
CBetaDriver::CBetaDriver( void )
{
  Ferr err;

  m_bInitPath = m_bInitChannel = false;
  m_iOutputMax = -1;
  m_iInputCh = m_iBufSize = m_iBufUse = 0;
  m_pBuf = NULL;

  m_iPrintMode = PMODE_NORMAL; // ʏ탂[hŒł悢

  // wK_Cȕ
  err = RCInit();
  if( err != FENOER )
    fprintf( stderr, "Error in RCInit(), err = %d\n", (int)err );
}


/*
  @brief      fXgN^
  @author     AdIn Research,Inc
  @date       2007/02/22
*/
CBetaDriver::~CBetaDriver(		void )
{
  clearBuf();

  // wK_Cǔ㏈
  RCHalt();
}


/*
  @brief      p^t@CAwKʃt@Ci[fBNgpXݒ肷
  @param[in]  sDir fBNgpX
  @return     bool
  @author     AdIn Research,Inc
  @date       2007/02/22
*/
bool CBetaDriver::SetPath( const char* sDir )
{
  char  sTmp[PATH_MAX];
  FILE* fp;

  // ̃fBNget@CpXݒ
  strcpy( sTmp, sDir );
  if( sTmp[ strlen( sTmp ) - 1 ] == '/' )
    sTmp[ strlen( sTmp ) - 1 ] = 0x00;
  sprintf( m_sParamPath, "%s/%s", sTmp, BD_PARAM_NAME );
  sprintf( m_sStudyPath, "%s/%s", sTmp, BD_STUDY_NAME );
  sprintf( m_sInputPath, "%s/%s", sTmp, BD_INPUT_NAME );

  // p^t@C̑݊mF
  fp = fopen( m_sParamPath, "rt" );
  if( ! fp ) {
    // p^t@C΁A쐬
    fp = fopen( m_sParamPath, "wt" );
    if( fp )
      fprintf( fp, BD_PARAM_VALUE );

    if( ! fp ) {
      fprintf( stderr, "Param file(%s) missing.\n", m_sParamPath );
      m_bInitPath = false;
      return false;
    }
  }
  fclose( fp );

  // wKʃt@C̑݊mF
  fp = fopen( m_sStudyPath, "rt" );
  // wKʃt@C݂ꍇ
  if( ! fp ) {
    if( errno != ENOENT ) {
      fprintf( stdout, "Error in fopen(%s), errno=%d.\n", m_sStudyPath, errno );
      m_bInitPath = false;
      return false;
    }
    // wKԃtOIɂ
    m_bStudyOnly = true;
  }
  // wKʃt@C݂Ȃꍇ
  else {
    // wKԃtOItɂ
    m_bStudyOnly = false;
    fclose( fp );
  }

  m_bInitPath = true;

  return true;
}


/*
  @brief      `liʂ̍ڐjݒ
  @param[in]  iCh `l
  @return     
  @author     AdIn Research,Inc
  @date       2007/02/22
*/
bool CBetaDriver::SetChannel( int iCh )
{
  // `lArŕύX̂͋֎~
  if( m_bInitChannel && m_iInputCh != iCh ) {
    fprintf( stderr, "Can not change input channel count.\n" );
    return false;
  }

  // `l̃`FbN
  if( iCh < 1 ) {
    fprintf( stderr, "Bad arg %d, to SetChannel().\n", iCh );
    return false;
  }

  // `lێ
  m_iInputCh = iCh;

  m_bInitChannel = true; // `lݒς݃tO
  return true;
}



/*
  @brief      wKE_s
  @param[in]  iOutputID IDԍi_[h́A-1wj
  @param[in]  pdData ʃf[^
  @param[out] piRet _ID iwK[h́AIDԍ̂܂܏ój
  @param[out] pdTrust vxiwK[h́A-1.0ój
  @return     bool
  @author     AdIn Research,Inc
  @date       2007/02/22
*/
bool CBetaDriver::DriveRna( int iOutputID, const double* pdData, int* piRet, double* pdTrust )
{
  tOneDatas tBuf;
  int       i;
  bool      bRet;
  Ferr      err;


  // KvȐݒ͂Ă邩mF
  if( ! m_bInitPath    /*fBNgpX͐ݒς݁H*/ || 
      ! m_bInitChannel /*`l͐ݒς݁H*/ ) {
    fprintf( stderr, "Not complete init process.\n" );
    return false;
  }

  // wKԂ̂ƂɁAIDԍݒ肳ĂȂꍇ́AG[
  if( m_bStudyOnly && iOutputID < 0 ) {
    fprintf( stderr, "Detect need rs1.bra, now study-only mode.\n" );
    return false;
  }

  // IDԍAőIDԍȏł΁AG[ƂiʑΏۂÂŁj
  if( iOutputID >= BD_OUTPUT_MAX ) {
    fprintf( stderr, "Too big outputID(%d), outputID < 8.\n", iOutputID );
    return false;
  }

  // IDԍAݓo^ĂID̍őԍ傫ꍇ̑΍
  if( m_iOutputMax < iOutputID )
    m_iOutputMax = iOutputID;


  // wKE_Cui-RNAj́Aʃf[^t@CŎ󂯕t邽߁A
  // ʃf[^AwKt@Cɕۑ
  {
    tBuf.pdData = (double*)malloc( sizeof( double ) * m_iInputCh );
    if( ! tBuf.pdData ) {
      fprintf( stderr, "error in malloc()\n" );
      return false;
    }
    // IDԍݒi_[h́A-1j
    tBuf.iOutputID = iOutputID;
    // ʃf[^ݒ
    for( i = 0; i < m_iInputCh; i++ )
      tBuf.pdData[i] = pdData[i];

    // ʃf[^wKt@Cɕۑ
    bRet = outDataFile( 1, &tBuf );
    free( (void*)tBuf.pdData );

    if( ! bRet ) {
      fprintf( stderr, "Error in create data file.\n" );
      return false;
    }
  }

  // wKE_Cui-RNAj̎s
  {
    // wKԂł͂ȂƂ
    if( ! m_bStudyOnly ) {
      // wKʃt@Cǂݍ
      err = RCLoadData( m_sStudyPath );
      if( err != FENOER ) {
        fprintf( stderr, "Error in RCLoadData(), err = %d\n", (int)err );
        return false;
      }
    }

    // wK[ĥƂ
    if( iOutputID >= 0 ) {
      // _ʂɁA_~[lݒ
      *piRet = iOutputID;
      *pdTrust = -1.0;
 
      // wKs
      if( ! study() ) {
        fprintf( stderr, "Error in RNA-study-process.\n" );
        return false;
      }
      // wKԃtOItɂ
      if( m_bStudyOnly )
        m_bStudyOnly = false;
    }
    // _[ĥƂ
    else {
      // _s
      if( ! detect( piRet, pdTrust ) ) {
        fprintf( stderr, "Error in RNA-detect-process.\n" );
        return false;
      }
    }
  }

  return true;
}


/*
  @brief      wKp̓ʃf[^AAwKpobt@ɒǉ
  @param[in]  iOutputID IDԍ
  @param[in]  pdData ʃf[^
  @param[in]  pdData
  @return     bool
  @author     AdIn Research,Inc
  @date       2007/02/22
*/
bool CBetaDriver::AddData( int iOutputID, const double* pdData )
{
  tOneDatas* pBuf;
  int        i;

  // KvȐݒ͂Ă邩mF
  if( ! m_bInitPath    /*fBNgpX͐ݒς݁H*/ ||
      ! m_bInitChannel /*`l͐ݒς݁H*/ ) {

    fprintf( stderr, "Not complete init process.\n" );
    return false;
  }

  // IDԍ`FbN
  {
    // IDԍAőIDԍȏł΁AG[ƂiʑΏۂÂŁj
    if( iOutputID >= BD_OUTPUT_MAX ) {
      fprintf( stderr, "Too big outputID(%d), outputID < 8.\n", iOutputID );
      return false;
    }
    if( iOutputID < 0 ) {
      fprintf( stderr, "Bad outputID(%d)\n", iOutputID );
      return false;
    }
  }

  // AwKpobt@g
  if( ! expandBuf() ) {
    fprintf( stderr, "error in expandBuf()\n" );
    return false;
  }

  // AwKpobt@ɋf[^ǉ
  {
    // ǉʒu擾
    pBuf = m_pBuf + m_iBufUse;
    pBuf->pdData = (double*)malloc( sizeof( double ) * m_iInputCh );
    if( ! pBuf->pdData ) {
      fprintf( stderr, "error in malloc()\n" );
      return false;
    }
    // IDԍݒ
    pBuf->iOutputID = iOutputID;

    // ʃf[^ݒ
    for( i = 0; i < m_iInputCh; i++ )
      pBuf->pdData[i] = pdData[i];
  }

  // IDԍőlXV
  if( m_iOutputMax < iOutputID )
    m_iOutputMax = iOutputID;

  // AwKpobt@ɂ܂f[^XV
  m_iBufUse++;

  return true;
}


/*
  @brief      AwKpobt@ɂ܂f[^gāAwKꊇs
  @return     bool
  @author     AdIn Research,Inc
  @date       2007/02/22
*/
bool CBetaDriver::DriveRna( void )
{
  bool bRet;
  Ferr err;


  // wKE_Cui-RNAj́Aʃf[^t@CŎ󂯕t邽߁A
  // AwKpobt@ɂ܂AwKp̓ʃf[^AwKt@Cɕۑ
  {
    // AwKpobt@ɂ܂f[^`FbN
    if( ! m_iBufUse ) {
      fprintf( stderr, "No data buffering\n" );
      return false;
    }
    // AwKpobt@ɂ܂AwKp̓ʃf[^AwKt@Cɏo
    bRet = outDataFile( m_iBufUse, m_pBuf );
    // AwKpobt@NA
    clearBuf();

    if( ! bRet ) {
      fprintf( stderr, "Error in create data file.\n" );
      fprintf( stderr, "Buffering data is deleted.\n" );
      return false;
    }
  }

  // wKԂł͂ȂƂ
  if( ! m_bStudyOnly ) {
    // wKʃt@Cǂݍ
    err = RCLoadData( m_sStudyPath );
    if( err != FENOER ) {
      fprintf( stderr, "Error in RCLoadData(), err = %d\n", (int)err );
      return false;
    }
  }
  // wKs
  if( ! study() ) {
    fprintf( stderr, "Error in RNA-study-process.\n" );
    return false;
  }
  // wKԃtOAItɂ
  if( m_bStudyOnly )
    m_bStudyOnly = false;

  return true;
}
