/******************************************************************************
 *      Copyright      Advanced Intelligence Research, Inc.   1988-1991       *
 ******************************************************************************
 ******************************************************************************
 *
 *  File name  : Kcommon.h
 *
 *  Note       : Header File for Common Use
 *
 * $Revision: 1.9 $
 * $Author: fujimaki $
 * $Date: 1994/03/01 08:36:17 $
 * $Locker:  $
 *
 ******************************************************************************/
#if !defined(KCOMMON_H)
#define KCOMMON_H

#include "Kstddef.h"            /* header for common definition */
#include "Kmemwat.h"            /* header for memory watch */

/******************** macros ********************/
#define	syserr(msg)         (syserr2(msg), exit(FAIL))
#define	fatal(msg, err)	    (fatal2((msg), (err)), exit(FAIL))

/******************** MS Windows, etc. ********************/
#if defined(MSWIN3)
# if defined(TEST_UNIX)
#   undef MSWIN3
# else   /* not TEST_UNIX */
#  if !defined(MSC6)
#    define MSC6            (1)
#  endif   /* not MSC6 */
# endif  /* not TEST_UNIX */
#endif   /* MSWIN3 */

#if defined(AIX3)
#  include <sys/utsname.h>
#endif /* AIX3 */

/******************** host information ********************/
#undef NO_HOSTID
#undef NO_HOSTNAME

#if defined(SOLARIS2)
# include <sys/systeminfo.h>
#endif /* SOLARIS2 */

#if defined(VMS)
# include <starlet.h>
# include <syidef.h>
#endif /* VMS */

#if defined(HPUXA0901) || defined(UNIOS_U) || defined(RTUX386)
# define NO_HOSTID
# define NO_HOSTNAME
#endif /* HPUXA0901 || UNIOS_U || RTUX386 */

#if defined(MS_DOS) || defined(AIX3)
# define NO_HOSTID
#endif /* MS_DOS || AIX3 */

#if defined(NO_HOSTNAME)
  extern int PROTOTYPE(gethostname, (char *, int));
# define L_HOSTNAME         (8)
#endif /* NO_HOSTNAME */

#if !defined(L_HOSTNAME)
# define L_HOSTNAME      (32)
#endif /* not L_HOSTNAME */

/*****************************************************************************
                          Authorization Code
 *****************************************************************************/
#define AUTH_LEN        (32)            /* maximum Acode length */
#if defined(VMS)
# define ENV_ACODE      "HOME"          /* environment name of Acode path */
#else /* not VMS */
# define ENV_ACODE      "ACODE_ADIN"    /* environment name of Acode path */
#endif /* not VMS */

#undef AUTH_FILE
#undef AUTH_MAGIC_NUMBER

#if defined(AIDNA1) || defined(AIDNA2)
  /* AI-DNA model01/02 */
# define AUTH_FILE          "acodeD1.AdIn"
# define AUTH_MAGIC_NUMBER  (0xa10d1991L)
# define AUTH_SALT          "DN"
#endif /* AIDNA1 || AIDNA2 */

#if defined AIDNAX
  /* AI-DNA modelX */
# define AUTH_FILE          "acodeDX.AdIn"
# define AUTH_MAGIC_NUMBER  (0xa30d1991L)
# define AUTH_SALT          "DN"
#endif /* AIDNAX */

#if defined(AIDNAXJ)
  /* AI-DNA modelX Jr. */
# define AUTH_FILE          "acodeDXJ.AdIn"
# define AUTH_MAGIC_NUMBER  (0xa31d1991L)
# define AUTH_SALT          "DN"
#endif /* AIDNAXJ */

#if defined(AIRNA1)
  /* AI-RNA model01 */
# define AUTH_FILE          "acodeR1.AdIn"
# define AUTH_MAGIC_NUMBER  (0xa10f1991L)
# define AUTH_SALT          "RN"
#endif /* AIRNA1 */

#if defined(AIRNA2)
  /* AI-RNA model02 */
# define AUTH_FILE          "acodeR2.AdIn"
# define AUTH_MAGIC_NUMBER  (0xa20f1991L)
# define AUTH_SALT          "RN"
#endif /* AIRNA2 */

#if defined(BRNA)
  /* Beta-RNA V1.2 */
# define AUTH_FILE          "acodeR.AdIn"
# define AUTH_MAGIC_NUMBER  (0xa30f1991L)
# define AUTH_SALT          "RN"
#endif /* BRNA */

#if !defined(AUTH_FILE)
  /* others */
# define AUTH_FILE          "acode.AdIn"
#endif /* not AUTH_FILE */

#if !defined(AUTH_MAGIC_NUMBER)
  /* others */
# define AUTH_MAGIC_NUMBER  (0xa01d1991L)
#endif /* not AUTH_MAGIC_NUMBER */

#if !defined(AUTH_SALT)
  /* others */
# define AUTH_SALT          "DN"
#endif /* AUTH_SALT */

/******************** function prototype ********************/
extern const char *PROTOTYPE(GetAdInCopyrightString, (void));
extern void  PROTOTYPE(syserr2, (char *));
extern void  PROTOTYPE(fatal2, (char *, int));
extern void  PROTOTYPE(KConvBitToChar,
                       (unsigned char *, unsigned char *, unsigned));
extern void  PROTOTYPE(KConvCharToBit,
                       (unsigned char *, unsigned char *, unsigned));
extern void  PROTOTYPE(KSwapByteStream,
                       (unsigned char *, unsigned));
extern int   PROTOTYPE(KCheckAcodeFile, (void));

#endif /* !KCOMMON_H */
