/******************************************************************************
 *      (C)Copyright        AdIn Research, Inc.         1991-1992.            *
 ******************************************************************************
 ******************************************************************************
 *
 *  File name  : rna00usr.h
 *
 *  Note       : Beta-RNA Library Common Definition File
 *
 * $Revision: 1.45 $
 * $Author: fujimaki $
 * $Date: 1999/11/16 06:43:12 $
 * $Locker:  $
 *
 *****************************************************************************/
#if !defined(_RNA00USR_H_)
#define _RNA00USR_H_

/**STDEXT_BEGIN**/
#if !defined(WIN32) && defined(_WIN32)
# define WIN32
#endif
#if defined(WIN32) && defined(_DLL)
# pragma pack(4)
#endif /* WIN32 and _DLL */
/**STDEXT_END**/

#if defined(WIN32)
# include "rnawin32.h"
#else /* not WIN32 */
# include "Kstddef.h"
#endif /* not WIN32 */

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

/**STDEXT_BEGIN**/
#undef EXPORT
#undef PUBLIC
#if defined(WIN32)
# if defined(_DLL)
#   define EXPORT   extern __declspec(dllexport)
#   define PUBLIC   __stdcall
# else /* not _DLL */
#   define PUBLIC   WINAPI
# endif /* not _DLL */
#endif /* not WIN32 */
#if !defined(PUBLIC)
# define PUBLIC
#endif /* not PUBLIC */
#if !defined(EXPORT)
# define EXPORT extern
#endif /* not EXPORT */
/**STDEXT_END**/

/********** Auxiliary Functions **********/
#if defined(VMS)
# define strNcpy        strncpy
#else /* not VMS */
EXPORT char *PROTOTYPE(strNcpy, (char *, const char *, int));
#endif /* not VMS */
EXPORT char *PROTOTYPE(strDup, (const char *));

/********** Initialization Macro **********/
#define RCINIT_M00  (RCInit00())    /* for model00 */
#define RCINIT_M01  (FENOER)        /* for model01 */
#define RCINIT_M02  (FENOER)        /* for model02 */
#define RCINIT_L00  (FENOER)        /* for learning model00 */
#define RCINIT_L01  (FENOER)        /* for learning model01 */
#define RCINIT_L02  (FENOER)        /* for learning model02 */
#define RCINIT_B00  (FENOER)        /* for label model00 */
/********** Halting Macro **********/
#define RCHALT_M00  (RCHalt00())    /* for model00 */
#define RCHALT_M01  (FENOER)        /* for model01 */
#define RCHALT_M02  (FENOER)        /* for model02 */
#define RCHALT_L00  (FENOER)        /* for learning model00 */
#define RCHALT_L01  (FENOER)        /* for learning model01 */
#define RCHALT_L02  (FENOER)        /* for learning model02 */
#define RCHALT_B00  (FENOER)        /* for label model00 */
/********** Masks **********/
#define RI_M00      (0x0001)        /* for model00 */
#define RI_M01      (0x0002)        /* for model01 */
#define RI_M02      (0x0004)        /* for model02 */
#define RI_L00      (0x0008)        /* for learning model00 */
#define RI_L01      (0x0010)        /* for learning model01 */
#define RI_L02      (0x0020)        /* for learning model02 */
#define RI_B00      (0x0040)        /* for label model00 */
#define RI_ALL      (0xffff)        /* all! */

#define RCInitMonitor()     (RCInitMode(0))
#define RCSetInitFlag()     (RCInitMode(1))
#define RCResetInitFlag()   (RCInitMode(2))

#define RCInit() \
    (((RCInitMonitor() != FENOER) && \
      ((RCINIT_M00) == FENOER) && \
      ((RCINIT_M01) == FENOER) && \
      ((RCINIT_M02) == FENOER) && \
      ((RCINIT_B00) == FENOER) && \
      ((RCINIT_L00) == FENOER) && \
      ((RCINIT_L01) == FENOER) && \
      ((RCINIT_L02) == FENOER) \
      ) ? (RCSetInitFlag()) : (F1ETCE))

#define RCHalt() \
    (((RCInitMonitor() == FENOER) && \
      ((RCHALT_L02) == FENOER) && \
      ((RCHALT_L01) == FENOER) && \
      ((RCHALT_L00) == FENOER) && \
      ((RCHALT_B00) == FENOER) && \
      ((RCHALT_M02) == FENOER) && \
      ((RCHALT_M01) == FENOER) && \
      ((RCHALT_M00) == FENOER) \
      ) ? (RCResetInitFlag()) : (F1ETCE))

/**STDEXT_BEGIN**/

/********** Basic Types **********/
typedef int                     RNUM;
typedef int                     RID;
#define UNDEFINED_ID            ((RID) -1)
#define MINIMUM_ID              ((RID) 0)
#define VALID_ID(rid)           ((rid) > (UNDEFINED_ID))
typedef unsigned long           RANA;
#define RANA_MIN                ((RANA) 0)
#define RANA_MAX                ((RANA) ULONG_MAX)
typedef unsigned char           RDIG;
#define ROFF                    ((RDIG) 0)
#define RON                     ((RDIG) 1)
typedef double                  RMD;
typedef enum {
    RB_FALSE = 0,
    RB_TRUE  = 1
} RBOOL;
/**STDEXT_END**/
typedef unsigned long           RCYCLE;

/**STDEXT_BEGIN**/

#define MATCH_DEGREE_MIN        ((RMD) 0.0)
#define MATCH_DEGREE_MAX        ((RMD) 100.0)
#define WEIGHT_MIN              (RANA_MIN)
#define WEIGHT_MAX              ((RANA) 0x0ff)
/**STDEXT_END**/
#define WEIGHT_BITS             (8)

/**STDEXT_BEGIN**/

#define DEGREE_EQUAL(ga, gb)    (fabs((ga) - (gb)) < DBL_MIN)
#define inBetween(n, a, b)      ((n) >= (a) && (n) <= (b))
/**STDEXT_END**/

/********** Membership Function (Analogue Pattern Data) **********/
typedef struct {
    RANA   raCenter;            /* center value */
    RANA   raLvar;              /* left variance */
    RANA   raRvar;              /* right variance */
    RANA   raVag;               /* vagueness */
    GENPTR pParam;              /* don't care */
} RMF;

/**STDEXT_BEGIN**/

/********** Matching Type **********/
typedef unsigned int RMTYPE;
#define RM_NOTYPE           (0x0000)
#define RM_MIN              (0x0001)
#define RM_MAX              (0x0002)
#define RM_ADDMEAN          (0x0004)
#define RM_MULMEAN          (0x0008)
#define RM_RULE             (0x0010)
#define RM_NOT              (0x0020)
#define RM_RAW              (0x0100)
#define RM_WEIGHT           (0x0200)
#define RM_LINEAR           (0x0400)
#define RM_SWINT            (0x1000)
#define RM_SWEXT            (0x2000)
#define RM_SRAW             (0x4000)
#define RM_SWEIGHT          (0x8000)
/**STDEXT_END**/

/********** Matching Mode **********/
typedef enum {
    RMM_ALL = 0,
    RMM_FAST
} RMMODE;

/********** Rule Table Cell **********/
typedef unsigned int RCELLTYPE;
#define RT_ATOM             (0x0001)
#define RT_CELL             (0x0002)
#define RT_ANALOGUE         (0x0010)
#define RT_DIGITAL          (0x0020)

#define RT_ATOM_P(p)    (((p)->rctype & RT_ATOM) && !((p)->rctype & RT_CELL))
#define RT_CELL_P(p)    (!((p)->rctype & RT_ATOM) && ((p)->rctype & RT_CELL))
#define RT_ANA_P(p)     ((p)->rctype & RT_ANALOGUE)
#define RT_DIG_P(p)     ((p)->rctype & RT_DIGITAL)

typedef struct ruleCell_ {
    RCELLTYPE rctype;               /* cell type */
    RANA      raWeight;             /* weight */
    union {
        RID rid;                        /* ID (RT_ANALOGUE, RT_DIGITAL) */
        struct {
            RMTYPE            rmtype;       /* operator */
            struct ruleCell_ *pChildren;    /* children */
            RNUM              rnChildren;   /* number of children */
        } cell;
    } item;
    struct ruleCell_ *pBrother;     /* brothers */
} RCELL;

/**STDEXT_BEGIN**/

/********** Error Code **********/
typedef enum {
    /* no error */
    FENOER = 0,
    /* F0 (warnings, info, etc.) */
    F0NONM, F0DUNM, F0RNGE, F0NMAC, F0DSPE,
    F0IOER, F0DBER, F0MALC, F0ETCE,
    /* F1 */
    F1NOER, F1NONM, F1DUNM, F1RNGE, F1NMAC,
    F1DSPE, F1IOER, F1DBER, F1MALC, F1ETCE,
    /* F2 */
    F2NOER, F2NONM, F2DUNM, F2RNGE, F2NMAC,
    F2DSPE, F2IOER, F2DBER, F2MALC, F2ETCE,
    /* F3 */
    F3NOER, F3NONM, F3DUNM, F3RNGE, F3NMAC,
    F3DSPE, F3IOER, F3DBER, F3MALC, F3ETCE,
    /* F4 */
    F4NOER, F4NONM, F4DUNM, F4RNGE, F4NMAC,
    F4DSPE, F4IOER, F4DBER, F4MALC, F4ETCE,
    /* F5 */
    F5NOER, F5NONM, F5DUNM, F5RNGE, F5NMAC,
    F5DSPE, F5IOER, F5DBER, F5MALC, F5ETCE,
    /* F6 */
    F6NOER, F6NONM, F6DUNM, F6RNGE, F6NMAC,
    F6DSPE, F6IOER, F6DBER, F6MALC, F6ETCE,
    /* F7 */
    F7NOER, F7NONM, F7DUNM, F7RNGE, F7NMAC,
    F7DSPE, F7IOER, F7DBER, F7MALC, F7ETCE,
    /* F8 */
    F8NOER, F8NONM, F8DUNM, F8RNGE, F8NMAC,
    F8DSPE, F8IOER, F8DBER, F8MALC, F8ETCE,
    /* F9 */
    F9NOER, F9NONM, F9DUNM, F9RNGE, F9NMAC,
    F9DSPE, F9IOER, F9DBER, F9MALC, F9ETCE,
    /* FA */
    FANOER, FANONM, FADUNM, FARNGE, FANMAC,
    FADSPE, FAIOER, FADBER, FAMALC, FAETCE,
    /* FB */
    FBNOER, FBNONM, FBDUNM, FBRNGE, FBNMAC,
    FBDSPE, FBIOER, FBDBER, FBMALC, FBETCE,
    /* FC */
    FCNOER, FCNONM, FCDUNM, FCRNGE, FCNMAC,
    FCDSPE, FCIOER, FCDBER, FCMALC, FCETCE,
    /* FD */
    FDNOER, FDNONM, FDDUNM, FDRNGE, FDNMAC,
    FDDSPE, FDIOER, FDDBER, FDMALC, FDETCE,
    /* FX */
    FXNOER, FXNONM, FXDUNM, FXRNGE, FXNMAC,
    FXDSPE, FXIOER, FXDBER, FXMALC, FXETCE
} Ferr;

#define WarningP(err)       ((err) > FENOER && (err) < F1NOER)
#define ErrorP(err)         ((err) >= F1NOER && (err) <= FXETCE)

/********** Error Specifications **********/
typedef enum {
    ERR_NOERR = 0,
    /***** RNA *****/
    ERR_BADPARAM,
    ERR_UNDEF,
    ERR_RANGE,
    ERR_NONEXIST,
    ERR_NOMATCH,
    ERR_DUPLICATE,
    ERR_INCOMPLETE,
    ERR_INIT,
    ERR_NOLINIT,
    /***** file I/O *****/
    ERR_FOPEN,
    ERR_FCLOSE,
    ERR_FCREAT,
    ERR_FWRITE,
    ERR_FREAD,
    ERR_INCLEVEL,
    ERR_STREAMPOS,
    ERR_EOF,
    /***** OS interface *****/
    ERR_MALLOC,
    /***** miscellaneous *****/
    ERR_BADDATA,
    ERR_NOPARAM,
    ERR_SYNTAX,
    ERR_MATH,
    ERR_SYSTEM,
    ERR_BUG
} RERR;

/********** Extended Error Information **********/
#define ERRMSG_LEN          (128)

typedef struct {
    Ferr ferr;                  /* general error code */
    RERR rerr;                  /* specific error code */
    char abyPool[ERRMSG_LEN];   /* miscellaneous information */
} RESPEC;
/**STDEXT_END**/

/********** System Information **********/
typedef struct {
    RNUM rnNtbl;                /* number of normalization tables */
    RNUM rnRule;                /* number of matching rules */
    /***** model01 *****/
    RNUM rnSset;                /* number of sensor sets */
    RNUM rnPgrp;                /* number of pattern groups */
    RNUM rnPtbl;                /* number of pattern tables */
    /***** model02 *****/
    RNUM rnCgrp;                /* number of condition group */
    RNUM rnCtbl;                /* number of condition table */
    RNUM rnDtbl;                /* number of data table */
} RSYSINFO;

/********** ID Table Management **********/
typedef unsigned long RCAT;
/***** reserved *****/
#define CA_SYSTEM   (0x80000000)    /* reserved for internal use */
/***** model00 *****/
#define CA_NTBL     (0x00000001)    /* normalization table */
#define CA_RULE     (0x00000002)    /* matching rule */
/***** model01 *****/
#define CA_SSET     (0x00000010)    /* sensor set */
#define CA_ASEN     (0x00000020)    /* analogue sensor */
#define CA_DSEN     (0x00000040)    /* digital sensor */
#define CA_PGRP     (0x00000080)    /* pattern group */
#define CA_PTBL     (0x00000100)    /* pattern table */
#define CA_PSET     (0x00000200)    /* pattern set */
#define CA_APAT     (0x00000400)    /* analogue pattern */
#define CA_DPAT     (0x00000800)    /* digital pattern */
/***** model02 *****/
#define CA_CGRP     (0x00001000)    /* condition group */
#define CA_CTBL     (0x00002000)    /* condition table */
#define CA_CSET     (0x00004000)    /* condition set */
#define CA_ACON     (0x00008000)    /* analogue condition */
#define CA_DCON     (0x00010000)    /* digital condition */
#define CA_DTBL     (0x00020000)    /* data table */
#define CA_DSET     (0x00040000)    /* data set */
#define CA_ADAT     (0x00080000)    /* analogue data */
#define CA_DDAT     (0x00100000)    /* digital data */

/********** Hooks **********/
typedef enum {
    RH_CREATE = 0,
    RH_REMOVE,
    RH_EXPAND,
    RH_SHRINK,
    RH_ACT_MAX
} RHOOK_ACT;

/********** Index Level **********/
#define RID_INDEX_NTBL          (0)
#define RID_INDEX_RULE          (0)
/***** model 01 *****/
#define RID_INDEX_SSET          (0)
#define RID_INDEX_SENS          (1)
#define RID_INDEX_PGRP          (0)
#define RID_INDEX_PTBL          (0)
#define RID_INDEX_PSET          (1)
#define RID_INDEX_PATN          (2)
/***** model02 *****/
#define RID_INDEX_CTBL          (0)
#define RID_INDEX_CSET          (1)
#define RID_INDEX_COND          (2)
#define RID_INDEX_DTBL          (0)
#define RID_INDEX_DSET          (1)
#define RID_INDEX_DATA          (2)
/***** declaration size *****/
#define RID_INDEX_MAX_DCL       (3)

/********** Error Redirection **********/
EXPORT FILE *pErrout_BRNA_;

/********** Function Prototypes **********/

EXPORT Ferr PUBLIC PROTOTYPE(RCInitMode, (int));
EXPORT Ferr PUBLIC PROTOTYPE(RCInit00, (void));
EXPORT Ferr PUBLIC PROTOTYPE(RCHalt00, (void));
EXPORT RERR PUBLIC PROTOTYPE(RCGetError, (RESPEC *));
EXPORT void PUBLIC PROTOTYPE(RCClearError, (void));
EXPORT Ferr PUBLIC PROTOTYPE(RCGetSystemInfo, (RSYSINFO *));
EXPORT Ferr PUBLIC PROTOTYPE(RCGetDataEntries, (RCAT, RNUM *));

EXPORT Ferr PUBLIC PROTOTYPE(RCGetIDList, (RCAT, RID *, RNUM));

EXPORT Ferr PUBLIC PROTOTYPE(RCCreateNormTable, (RNUM, double *, RANA *, RID *));
EXPORT Ferr PUBLIC PROTOTYPE(RCDeleteNormTable, (RID));
EXPORT Ferr PUBLIC PROTOTYPE(RCGetNormTableInfo, (RID, RNUM *));
EXPORT Ferr PUBLIC PROTOTYPE(RCGetNormTableData, (RID, double *, RANA *));
EXPORT Ferr PUBLIC PROTOTYPE(RCRealToNorm, (RID, double, RANA *));
EXPORT Ferr PUBLIC PROTOTYPE(RCNormToReal, (RID, RANA, double *));

EXPORT Ferr PUBLIC PROTOTYPE(RCCreateMatchRule, (RCELL *, RID *));
EXPORT Ferr PUBLIC PROTOTYPE(RCDeleteMatchRule, (RID));
EXPORT Ferr PUBLIC PROTOTYPE(RCGetMatchRule, (RID, RCELL **));
EXPORT Ferr PUBLIC PROTOTYPE(RCFreeMatchRule, (RCELL *));

EXPORT Ferr PUBLIC PROTOTYPE(RCSetHook, (RCAT, RHOOK_ACT, Ferr (*)(RID *, RNUM)));
EXPORT Ferr PUBLIC PROTOTYPE(RCRemoveHook, (RCAT, RHOOK_ACT, Ferr (*)(RID *, RNUM)));

#if defined(__cplusplus)
}
#endif /* __cplusplus */

#endif /* not _RNA00USR_H_ */
