/******************************************************************************
 *      (C)Copyright        AdIn Research, Inc.         1991-1992.            *
 ******************************************************************************
 ******************************************************************************
 *
 *  File name  : rna02def.h
 *
 *  Note       : Beta-RNA model02 Developer's File
 *
 * $Revision: 1.7 $
 * $Author: fujimaki $
 * $Date: 1999/11/05 03:03:48 $
 * $Locker:  $
 *
 *****************************************************************************/
#if !defined(_RNA02DEF_H_)
#define _RNA02DEF_H_

#include "rna00def.h"
#include "rna02usr.h"

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

/********** Data Set **********/
typedef struct dset_ {
    RANA         *praAdat;      /* number of analogue data */
    RDIG         *prdDdat;      /* number of digital data */
    RMD          *prgEach;      /* matching degree of each data */
    RMD           rgTotal;      /* matching degree of this data set */
    RID           riCset;       /* best matched condition set ID */
    struct dset_ *pPrev;        /* pointer to the previous data set */
    struct dset_ *pNext;        /* pointer to the next data set */
} DSET;

/********** Data Table **********/
typedef struct dtbl_ {
    RNUM   rnDset;              /* number of data sets */
    RNUM   rnAdat;              /* number of analogue data */
    RNUM   rnDdat;              /* number of digital data */
    RID    riCtbl;              /* best matched condition table ID */
    DSET  *pDset;               /* pointer to the first data set */
    RID   *priDset;             /* data set ID array */
    RID   *priNtbl;             /* array of normalization table ID */
    RCYCLE rcycle;              /* counter */
} DTBL;

/********** Condition Set **********/
typedef struct cset_ {
    RMF          *prmAcond;     /* analogue condition (membership function) */
    RDIG         *prdDcond;     /* digital condition */
    RANA         *praWeight;    /* weight of each condition */
    RANA          raWeight;     /* weight */
    struct cset_ *pPrev;        /* pointer to the previous condition set */
    struct cset_ *pNext;        /* pointer to the next condition set */
} CSET;

/********** Condition Table **********/
typedef struct ctbl_ {
    RNUM   rnCset;              /* number of condition sets */
    RNUM   rnAcond;             /* number of analogue condition */
    RNUM   rnDcond;             /* number of digital condition */
    CSET  *pCset;               /* pointer to the first condition set */
    RMTYPE rmtype;              /* matching type */
    RID    riRule;              /* matching rule ID */
    RCYCLE rcycle;              /* counter */
} CTBL;

/********** Function Prototypes **********/
extern Ferr PROTOTYPE(rcGetCondSet, (CSET *, RID, CSET **));

#if defined(__cplusplus)
}
#endif /* __cplusplus */

#endif /* not _RNA02DEF_H_ */
