/******************************************************************************
 *      (C)Copyright        AdIn Research, Inc.         1991-1992.            *
 ******************************************************************************
 ******************************************************************************
 *
 *  File name  : rnaF0usr.h
 *
 *  Note       : Beta-RNA Data File Loader Application Programmer's File
 *
 * $Revision: 1.28 $
 * $Author: fujimaki $
 * $Date: 1999/11/05 03:03:48 $
 * $Locker:  $
 *
 *****************************************************************************/
#if !defined(_RNAF00USR_H_)
#define _RNAF00USR_H_

#include "rna00usr.h"           /* default */

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

/********** Generic Buffer Size **********/
#define FIO_GENERIC_BUFSIZ      (512)

/********** Stream **********/
typedef struct {
    FILE *pFile;
    char  sFile[MAXPATHLEN];
    int   nLine;
    RBOOL rbAlive;
    char *sBuf;
    int   nBufLen;
} STREAM;

/********** Model00 **********/
EXPORT Ferr PUBLIC PROTOTYPE(rcLoadM00Data,    (STREAM *));
EXPORT Ferr PUBLIC PROTOTYPE(rcSaveM00Data,    (STREAM *));
EXPORT Ferr PUBLIC PROTOTYPE(rcReadNormTbl,    (STREAM *));
EXPORT Ferr PUBLIC PROTOTYPE(rcReadMatchRule,  (STREAM *));
EXPORT Ferr PUBLIC PROTOTYPE(rcWriteNormTbl,   (STREAM *, RNUM));
EXPORT Ferr PUBLIC PROTOTYPE(rcWriteMatchRule, (STREAM *, RNUM));
EXPORT Ferr PUBLIC PROTOTYPE(rcReadMatchRuleMMI, (char *, RCELL **)); /* for MMI */
EXPORT Ferr PUBLIC PROTOTYPE(rcWriteMatchRuleMMI, (char *, RCELL *)); /* for MMI */
#define LOAD_M00                (rcLoadM00Data)
#define SAVE_M00                (rcSaveM00Data)
#define LOAD_M00_NORM_TABLE     (rcReadNormTbl)
#define LOAD_M00_MATCH_RULE     (rcReadMatchRule)
#define SAVE_M00_NORM_TABLE     (rcWriteNormTbl)
#define SAVE_M00_MATCH_RULE     (rcWriteMatchRule)

/********** Model01 **********/
#if defined(_RNA01USR_H_)
  EXPORT Ferr PUBLIC PROTOTYPE(rcLoadM01Data,       (STREAM *));
  EXPORT Ferr PUBLIC PROTOTYPE(rcSaveM01Data,       (STREAM *));
  EXPORT Ferr PUBLIC PROTOTYPE(rcReadPatternGroup,  (STREAM *));
  EXPORT Ferr PUBLIC PROTOTYPE(rcReadPatternTable,  (STREAM *));
  EXPORT Ferr PUBLIC PROTOTYPE(rcReadSensorSet,     (STREAM *));
  EXPORT Ferr PUBLIC PROTOTYPE(rcWritePatternGroup, (STREAM *, RNUM));
  EXPORT Ferr PUBLIC PROTOTYPE(rcWritePatternTable, (STREAM *, RNUM));
  EXPORT Ferr PUBLIC PROTOTYPE(rcWriteSensorSet,    (STREAM *, RNUM));
# define LOAD_M01               (rcLoadM01Data)
# define SAVE_M01               (rcSaveM01Data)
# define LOAD_M01_PATTERN_GROUP (rcReadPatternGroup)
# define LOAD_M01_PATTERN_TABLE (rcReadPatternTable)
# define LOAD_M01_SENSOR_SET    (rcReadSensorSet)
# define SAVE_M01_PATTERN_GROUP (rcWritePatternGroup)
# define SAVE_M01_PATTERN_TABLE (rcWritePatternTable)
# define SAVE_M01_SENSOR_SET    (rcWriteSensorSet)
#else /* !_RNA01USR_H_ */
# define LOAD_M01               (NULL)
# define SAVE_M01               (NULL)
# define LOAD_M01_PATTERN_GROUP (NULL)
# define LOAD_M01_PATTERN_TABLE (NULL)
# define LOAD_M01_SENSOR_SET    (NULL)
# define SAVE_M01_PATTERN_GROUP (NULL)
# define SAVE_M01_PATTERN_TABLE (NULL)
# define SAVE_M01_SENSOR_SET    (NULL)
#endif /* !_RNA01USR_H_ */

/********** Model02 **********/
#if defined(_RNA02USR_H_)
  EXPORT Ferr PUBLIC PROTOTYPE(rcLoadM02Data,    (STREAM *));
  EXPORT Ferr PUBLIC PROTOTYPE(rcSaveM02Data,    (STREAM *));
  EXPORT Ferr PUBLIC PROTOTYPE(rcReadDataTable,  (STREAM *));
  EXPORT Ferr PUBLIC PROTOTYPE(rcReadCondTable,  (STREAM *));
  EXPORT Ferr PUBLIC PROTOTYPE(rcWriteDataTable, (STREAM *, RNUM));
  EXPORT Ferr PUBLIC PROTOTYPE(rcWriteCondTable, (STREAM *, RNUM));
# define LOAD_M02               (rcLoadM02Data)
# define SAVE_M02               (rcSaveM02Data)
# define LOAD_M02_DATA_TABLE    (rcReadDataTable)
# define LOAD_M02_COND_TABLE    (rcReadCondTable)
# define SAVE_M02_DATA_TABLE    (rcWriteDataTable)
# define SAVE_M02_COND_TABLE    (rcWriteCondTable)
#else
# define LOAD_M02               (NULL)
# define SAVE_M02               (NULL)
# define LOAD_M02_DATA_TABLE    (NULL)
# define LOAD_M02_COND_TABLE    (NULL)
# define SAVE_M02_DATA_TABLE    (NULL)
# define SAVE_M02_COND_TABLE    (NULL)
#endif /* _RNA02USR_H_ */

/********** Label Model00 **********/
#if defined(_RNAB0USR_H_)
  EXPORT Ferr PUBLIC PROTOTYPE(rcSetupB0Fns, (void));
# define SETUP_B0_FNS           (rcSetupB0Fns())
#else /* !_RNAB0USR_H_ */
# define RLABEL_MAXLEN          (128)       /* for compatibility */
# define SETUP_B0_FNS           (FENOER)
#endif /* !_RNAB0USR_H_ */

/********** Learning Model00 **********/
#if defined(_RNAL0USR_H_)
  EXPORT Ferr PUBLIC PROTOTYPE(rcSetupL0Fns, (void));
  EXPORT Ferr PUBLIC PROTOTYPE(rcClearL0Fns, (void));
# define SETUP_L0_FNS           (rcSetupL0Fns())
#else /* !_RNAL0USR_H_ */
# define SETUP_L0_FNS           (FENOER)
#endif /* _RNAL0USR_H_ */

/********** Lexicons **********/
/***** I/O directives *****/
#define STR_INCLUDE_NAM             "@include"
/***** static values *****/
#define STR_DIGON_NAM               "ON"
#define STR_DIGOFF_NAM              "OFF"
#define STR_MAXNORM_NAM             "MAXNN"
/***** normalization *****/
#define STR_NTBL_BLK                "Normalization"
#define STR_IN_BLK                  "I"
#define STR_OUT_BLK                 "O"
/***** matching *****/
#define STR_TYPE_BLK                "Type"
#define STR_RULE_BLK                "MatchRule"
#define STR_MIN_NAM                 "Min"
#define STR_ADDMEAN_NAM             "AddMean"
#define STR_MULMEAN_NAM             "MulMean"
#define STR_RULE_NAM                "Rule"
#define STR_LINEAR_NAM              "Linear"
#define STR_RAW_NAM                 "Raw"
#define STR_WEIGHT_NAM              "Weight"
#define STR_SRAW_NAM                "SRaw"
#define STR_SWEIGHT_NAM             "SWeight"
#define STR_SWINT_NAM               "SWInt"
#define STR_SWEXT_NAM               "SWExt"
#define STR_ADD_OPR                 "Add"
#define STR_WADD_OPR                "WAdd"
#define STR_MULT_OPR                "Mul"
#define STR_WMULT_OPR               "WMul"
#define STR_NOT_OPR                 "Not"
/***** sensor set *****/
#define STR_SSET_BLK                "SensorSet"
#define STR_ANA_BLK                 "Analogue"
#define STR_DIG_BLK                 "Digital"
/***** pattern table *****/
#define STR_PTBL_BLK                "PatternTable"
#define STR_PSET_BLK                "PatternSet"
#define STR_MF_BLK                  "F"
#define STR_DVAL_BLK                "D"
#define STR_OUTSSET_BLK             "OutputSensor"
#define STR_OUTCOND_BLK             "OutputCond"
/***** pattern group *****/
#define STR_PGRP_BLK                "PatternGroup"
#define STR_FUSION_BLK              "Fusion"
#define STR_THRESHOLD_BLK           "Threshold"
/***** learning *****/
#define STR_LBASIC_BLK              "LearnBasic"
#define STR_LEXTEND_BLK             "LearnExtend"
#define STR_LWEIGHT_BLK             "LearnWeight"
#define STR_LSTATUS_BLK             "LearnStatus"
#define STR_LHIST_BLK               "LearnHistory"
#define STR_ACTIVE_NAM              "Active"
#define STR_HISTORY_NAM             "History"
#define STR_LEARN_NAM               "Learn"
#define STR_NOCLRHST_NAM            "NoClearHist"
#define STR_VAREXT_NAM              "VarExt"
#define STR_VARFIX_NAM              "VarFix"
#define STR_EXTEND_NAM              "Extend"
#define STR_SMALL_NAM               "Small"
#define STR_COMPENS_NAM             "Compensation"
#define STR_NOINHERIT_NAM           "NoInherit"
#define STR_ADDVAR_NAM              "AddVar"
/********** data table **********/
#define STR_DTBL_BLK                "DataTable"
#define STR_DSET_BLK                "DataSet"
/********** condition table **********/
#define STR_CTBL_BLK                "ConditionTable"
#define STR_CSET_BLK                "ConditionSet"

/********** I/O Main Entry **********/
typedef enum {
    OM_READ,
    OM_WRITE,
    OM_APPEND
} FOPEN_MODE;

#define FIO_MAIN_ARGLIST \
    Ferr (PUBLIC*)(STREAM *), \
    Ferr (PUBLIC*)(STREAM *), \
    Ferr (PUBLIC*)(STREAM *)

EXPORT Ferr PUBLIC PROTOTYPE(rcIOData, (char *, FOPEN_MODE, FIO_MAIN_ARGLIST));
EXPORT Ferr PUBLIC PROTOTYPE(rcIOSubData, (char *, FOPEN_MODE, unsigned, FIO_MAIN_ARGLIST));
EXPORT Ferr PUBLIC PROTOTYPE(rcLoadCatData, (char *, char *, Ferr (PUBLIC *)(STREAM *)));
EXPORT Ferr PUBLIC PROTOTYPE(rcSaveCatData, (char *, RCAT,
                                             Ferr (PUBLIC *)(STREAM *, RNUM),
                                             Ferr (PUBLIC *)(STREAM *, RNUM),
                                             Ferr (PUBLIC *)(STREAM *, RNUM),
                                             Ferr (PUBLIC *)(STREAM *, RNUM),
                                             Ferr (PUBLIC *)(STREAM *, RNUM),
                                             Ferr (PUBLIC *)(STREAM *, RNUM),
                                             Ferr (PUBLIC *)(STREAM *, RNUM)));

/********** Load Data **********/
#define RCLoadData(sFile) \
    ((SETUP_B0_FNS == FENOER \
      && SETUP_L0_FNS == FENOER \
      && ((RCInitMonitor() == FENOER) \
          ? (RCHalt()) : FENOER) == FENOER && RCInit() == FENOER) \
     ? rcIOData(sFile, OM_READ, LOAD_M00, LOAD_M01, LOAD_M02) \
     : F1ETCE)

#define RCLoadSubData(sFile, mask) \
    ((SETUP_B0_FNS == FENOER && SETUP_L0_FNS == FENOER) \
     ? rcIOSubData(sFile, OM_READ, mask, LOAD_M00, LOAD_M01, LOAD_M02) \
     : F1ETCE)

#define RCLoadCat(sFile, cat) \
    ((SETUP_B0_FNS == FENOER && SETUP_L0_FNS == FENOER) \
     ? (((((cat) & CA_NTBL) && rcLoadCatData((sFile), \
                                             STR_NTBL_BLK, LOAD_M00_NORM_TABLE) \
          != FENOER) \
         || (((cat) & CA_RULE) \
             && rcLoadCatData((sFile), STR_RULE_BLK, LOAD_M00_MATCH_RULE) \
             != FENOER) \
         || (((cat) & CA_PGRP) \
             && rcLoadCatData((sFile), STR_PGRP_BLK, LOAD_M01_PATTERN_GROUP) \
             != FENOER) \
         || (((cat) & CA_PTBL) \
             && rcLoadCatData((sFile), STR_PTBL_BLK, LOAD_M01_PATTERN_TABLE) \
             != FENOER) \
         || (((cat) & CA_SSET) \
             && rcLoadCatData((sFile), STR_SSET_BLK, LOAD_M01_SENSOR_SET) \
             != FENOER) \
         || (((cat) & CA_DTBL) \
             && rcLoadCatData((sFile), STR_DTBL_BLK, LOAD_M02_DATA_TABLE) \
             != FENOER) \
         || (((cat) & CA_CTBL) \
             && rcLoadCatData((sFile), STR_CTBL_BLK, LOAD_M02_COND_TABLE) \
             != FENOER)) ? F1ETCE : FENOER) \
     : F1ETCE)

/********** Save Data **********/
#define RCSaveData(sFile) \
    ((SETUP_B0_FNS == FENOER && SETUP_L0_FNS == FENOER) \
     ? rcIOData(sFile, OM_WRITE, SAVE_M00, SAVE_M01, SAVE_M02) \
     : F1ETCE)

#define RCSaveSubData(sFile, sMode, mask) \
    ((SETUP_B0_FNS == FENOER && SETUP_L0_FNS == FENOER) \
     ? rcIOSubData(sFile, OM_WRITE, mask, SAVE_M00, SAVE_M01, SAVE_M02) \
     : F1ETCE)

#define RCSaveCat(sFile, cat) \
    ((SETUP_B0_FNS == FENOER && SETUP_L0_FNS == FENOER) \
     ? (rcSaveCatData((sFile), (cat), \
                      SAVE_M00_NORM_TABLE, \
                      SAVE_M00_MATCH_RULE, \
                      SAVE_M01_SENSOR_SET, \
                      SAVE_M01_PATTERN_TABLE, \
                      SAVE_M01_PATTERN_GROUP, \
                      SAVE_M02_DATA_TABLE, \
                      SAVE_M02_COND_TABLE)) \
     : F1ETCE)

#if defined(__cplusplus)
}
#endif /* __cplusplus */

#endif /* not _RNAF00USR_H_ */
