/******************************************************************************
 *      Copyright      Advanced Intelligence Research, Inc.   1991-1992       *
 ******************************************************************************
 ******************************************************************************
 *
 *  File name  : rnaS0usr.h
 *
 *  Note       : Beta-RNA Simulator (common)
 *
 * $Revision: 2.22 $
 * $Author: fujimaki $
 * $Date: 1999/11/16 06:34:31 $
 * $Locker:  $
 *
 *****************************************************************************/
#if !defined(RNAS0USR_H)
#define RNAS0USR_H

#include "rnaL0usr.h"
#include "rnaF0usr.h"

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

/**STDEXT_BEGIN**/
/*---------------------------- Simulator Type (P.G.) ------------------------*/
typedef enum {
    RST_NON = 0, RST_RS1, RST_RS2_IN, RST_RS2_OUT
} RSTYPE;

/*---------------------------- ulMode of Simulator --------------------------*/
/*                  these mode must be exclusive from RLMM_* (in rnaL0usr.h) */
#define MRS_ADD     (0x01000000L) /* Calc.Total M.Value
                                          (on->Add Mean, off->Times Mean)    */
#define MRS_LEVEL   (0x02000000L) /* Use Level(P.G.), off->Not Use [rs2]     */
#define MRS_SWEIGHT (0x04000000L) /* Use Pattern Set Weight                  */
#define MRS_NO_ACT  (0x08000000L) /* No Check Act/Inact of MF & PS           */
#define MRS_SELECT  (0x10000000L) /* Select Sensor Mode                      */
#define MRS_LINEAR  (0x20000000L) /* RS2 Linaerity Mode                      */
#define MRS_MINSQR  (0x40000000L) /* Compensation with Minimum Square Root   */

/*---------------------------- iOutMode of RS2DATA --------------------------*/
#define MORS2_APWAM (0x00) /* Arithmetical Progression (Weight Add Mean)     */
#define MORS2_APT3W (0x01) /* Arithmetical Progression (Top & Near WAM.)     */
#define MORS2_APT2W (0x02) /* Arithmetical Progression (Top & Near Eff.)     */
#define MORS2_PSWAM (0x10) /* PS Sort Result (weight = matching degree)      */
#define MORS2_PSWWM (0x11) /* PS Sort Result (weight = MD * Out-PS's weight) */

/*--------------------------- Function Prototype (stdext) -------------------*/
#define RS_SetLParaDefaults(rstype,pulStat,prlm,prle,prlw,praSysMax) \
        (rsSetLParaDefaults(rstype,pulStat,prlm,prle,prlw,praSysMax))
#define RS_LoadLParaFile(sF,iO,nA,uS,uM,lm,le,lw,uSO,uMO,lmO,leO,lwO,rS,m2thr) \
        (rsLoadLParaFile(sF,iO,nA,uS,uM,lm,le,lw,uSO,uMO,lmO,leO,lwO,rS,m2thr))
#define RS_SaveLParaFile(sF,iO,nA,uS,uM,lm,le,lw,uSO,uMO,lmO,leO,lwO,rS,m2thr) \
        (rsSaveLParaFile(sF,iO,nA,uS,uM,lm,le,lw,uSO,uMO,lmO,leO,lwO,rS,m2thr))

EXPORT void PUBLIC PROTOTYPE( rsSetLParaDefaults,
              (RSTYPE, unsigned long *, RLMFP *, RLEXT *, RLWEIGHT *, RANA *));
EXPORT Ferr PUBLIC PROTOTYPE( rsLoadLParaFile,
              (const char*, int *, RNUM *,
               unsigned long *, unsigned long *, RLMFP *, RLEXT *, RLWEIGHT *,
               unsigned long *, unsigned long *, RLMFP *, RLEXT *, RLWEIGHT *,
               RANA *, RMD*));          /* 96.02.06 */
EXPORT Ferr PUBLIC PROTOTYPE( rsSaveLParaFile,
              (const char*, int, RNUM,
               unsigned long, unsigned long, RLMFP *, RLEXT *, RLWEIGHT *,
               unsigned long, unsigned long, RLMFP *, RLEXT *, RLWEIGHT *,
               RANA, RMD));
EXPORT Ferr PUBLIC PROTOTYPE( RS_SetInput, (RID, RNUM, double *));
/**STDEXT_END**/

/*----------------------------------- Print Mode ----------------------------*/
#define  PMODE_NORMAL  (0x0001)
#define  PMODE_VERBOSE (0x0002)
#define  PMODE_SIMPLE  (0x0004)

/*--------------------------- Function Prototype ----------------------------*/
EXPORT Ferr PROTOTYPE( rsOpenInputFile, (const char *, STREAM *));
EXPORT Ferr PROTOTYPE( rsCloseInputFile, (STREAM *));
EXPORT Ferr PROTOTYPE( rsLoadData,
                      (STREAM *, RNUM, RNUM, RID, int *, double *, int));
EXPORT Ferr PROTOTYPE( rsLoadData2, (STREAM *, RNUM, RID));

#if defined(__cplusplus)
}
#endif /* __cplusplus */

#endif /* RNAS0USR_H */
